# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

from typing import Any


def args_kwargs(obj: Any) -> bool:
    if obj is None:
        return False

    try:
        args, kwargs = obj
    except (ValueError, TypeError):
        return False

    if not isinstance(args, (list, tuple)):
        return False

    if not isinstance(kwargs, dict):
        return False

    return True



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

from typing import Any, Callable, Dict, List

from . import callable_extenders, module_extension


class Registry:
    def __init__(self) -> None:
        self._modules_extensions: Dict[str, module_extension.ModuleExtension] = {}

    @property
    def module_names(self):  # type: ignore
        return self._modules_extensions.keys()

    def get_extension(self, module_name: str) -> module_extension.ModuleExtension:
        return self._modules_extensions[module_name]

    def _get_callable_extenders(
        self, module_name: str, callable_name: str
    ) -> callable_extenders.CallableExtenders:
        extension = self._modules_extensions.setdefault(
            module_name, module_extension.ModuleExtension()
        )
        return extension.extenders(callable_name)

    def register_before(
        self, module_name: str, callable_name: str, patcher_function: Callable
    ) -> None:
        """
        patcher_function: Callable with the following signature
            func(
                original,
                *args,
                **kwargs
            )
        original - original callable to patch

        Return value of patcher function is expected to be either None
        or [Args,Kwargs] tuple to overwrite original args and kwargs
        """
        extenders = self._get_callable_extenders(module_name, callable_name)
        extenders.before.append(patcher_function)

    def register_after(
        self, module_name: str, callable_name: str, patcher_function: Callable
    ) -> None:
        """
        patcher_function: Callable with the following signature
            func(
                original,
                return_value,
                *args,
                **kwargs
            )
        original - original callable to patch
        return_value - value returned by original callable

        Return value of patcher function will overwrite return_value of
        patched function if not None
        """
        extenders = self._get_callable_extenders(module_name, callable_name)
        extenders.after.append(patcher_function)

    def register_after_exception(
        self, module_name: str, callable_name: str, patcher_function: Callable
    ) -> None:
        """
        patcher_function: Callable with the following signature
            func(
                original,
                exception,
                *args,
                **kwargs
            )
        original - original callable to patch
        exception - exception thrown from original callable

        Expected to return None.
        """
        extenders = self._get_callable_extenders(module_name, callable_name)
        extenders.after_exception.append(patcher_function)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import io
import json
from typing import Dict, Optional

import comet_ml

from .. import convert
from ..chains import deepmerge
from ..types import JSONEncodable


class LLMTraceAPI:
    _api_experiment: comet_ml.APIExperiment

    def __init__(self) -> None:
        raise NotImplementedError(
            "Please use API.get_llm_trace_by_key or API.get_llm_trace_by_name methods to get the instance"
        )

    @classmethod
    def __api__from_api_experiment__(
        cls, api_experiment: comet_ml.APIExperiment
    ) -> "LLMTraceAPI":
        instance = object.__new__(cls)
        instance._api_experiment = api_experiment

        return instance

    def get_name(self) -> Optional[str]:
        """
        Get the name of the trace
        """
        return self._api_experiment.get_name()  # type: ignore

    def get_key(self) -> str:
        """
        Get the unique identifier for this trace
        """
        return self._api_experiment.key  # type: ignore

    def log_user_feedback(self, score: float) -> None:
        """
        Log user feedback

        Args:
            score: float, the feedback score. Can be either 0, 0.0, 1 or 1.0
        """
        ALLOWED_SCORES = [0.0, 1.0]
        if score not in ALLOWED_SCORES:
            raise ValueError(
                f"Score it not valid, should be {ALLOWED_SCORES} but got {score}"
            )

        self._api_experiment.log_metric("user_feedback", score)

    def _get_trace_data(self) -> Dict[str, JSONEncodable]:
        try:
            asset_id = next(
                asset
                for asset in self._api_experiment.get_asset_list()
                if asset["fileName"] == "comet_llm_data.json"
            )["assetId"]
        except Exception as exception:
            raise ValueError(
                "Failed update metadata for this trace, metadata is not available"
            ) from exception

        trace_data = json.loads(self._api_experiment.get_asset(asset_id))

        return trace_data  # type: ignore

    def get_metadata(self) -> Dict[str, JSONEncodable]:
        """
        Get trace metadata
        """
        trace_data = self._get_trace_data()

        return trace_data["metadata"]  # type: ignore

    def log_metadata(self, metadata: Dict[str, JSONEncodable]) -> None:
        """
        Update the metadata field for a trace, can be used to set or update metadata fields

        Args:
            metadata_dict: dict, dict in the form of {"metadata_name": value, ...}. Nested metadata is supported
        """

        trace_data = self._get_trace_data()
        updated_trace_metadata = deepmerge.deepmerge(
            trace_data.get("metadata", {}), metadata
        )
        trace_data["metadata"] = updated_trace_metadata

        stream = io.StringIO()
        json.dump(trace_data, stream)
        stream.seek(0)
        self._api_experiment.log_asset(
            stream, overwrite=True, name="comet_llm_data.json"
        )

        self._api_experiment.log_parameters(
            convert.chain_metadata_to_flat_parameters(metadata)
        )



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import collections
from typing import TYPE_CHECKING, Dict, List, Optional

from .. import datetimes
from ..types import JSONEncodable
from . import context, deepmerge, version

if TYPE_CHECKING:  # pragma: no cover
    from ..experiment_info import ExperimentInfo
    from . import span


class Chain:
    def __init__(
        self,
        inputs: JSONEncodable,
        metadata: Optional[Dict[str, JSONEncodable]],
        experiment_info: "ExperimentInfo",
        tags: Optional[List[str]] = None,
        others: Optional[Dict[str, JSONEncodable]] = None,
    ):
        self._nodes: List["span.Span"] = []
        self._node_names_registry: collections.defaultdict = collections.defaultdict(
            lambda: 0
        )
        self._inputs = inputs
        self._outputs: Optional[Dict[str, JSONEncodable]] = None
        self._metadata = metadata if metadata is not None else {}
        self._context = context.Context()
        self._prepare_timer()

        self._experiment_info = experiment_info
        self._tags = tags
        self._others = others if others is not None else {}

    @property
    def experiment_info(self) -> "ExperimentInfo":  # pragma: no cover
        return self._experiment_info

    @property
    def tags(self) -> Optional[List[str]]:
        return self._tags

    @property
    def others(self) -> Dict[str, JSONEncodable]:
        return self._others

    @property
    def context(self) -> "context.Context":  # pragma: no cover
        return self._context

    def _prepare_timer(self) -> None:
        self._timer = datetimes.Timer()
        self._timer.start()

    def track_node(self, node: "span.Span") -> None:
        self._nodes.append(node)

    def set_outputs(
        self,
        outputs: Dict[str, JSONEncodable],
        metadata: Optional[Dict[str, JSONEncodable]] = None,
    ) -> None:
        self._timer.stop()
        self._outputs = outputs

        if metadata is not None:
            self._metadata = deepmerge.deepmerge(self._metadata, metadata)

    def as_dict(self) -> Dict[str, JSONEncodable]:
        chain_nodes = [chain_node.as_dict() for chain_node in self._nodes]

        inputs = self._inputs
        outputs = self._outputs

        inputs = inputs if isinstance(inputs, dict) else {"input": inputs}
        outputs = outputs if isinstance(outputs, dict) else {"output": outputs}

        result = {
            "version": version.ASSET_FORMAT_VERSION,
            "chain_nodes": chain_nodes,
            "chain_inputs": inputs,
            "chain_outputs": outputs,
            "metadata": self._metadata,
            "category": "chain",
            "start_timestamp": self._timer.start_timestamp,
            "end_timestamp": self._timer.end_timestamp,
            "chain_duration": self._timer.duration,
        }

        return result

    def generate_node_name(self, category: str) -> str:
        name = f"{category}-{self._node_names_registry[category]}"
        self._node_names_registry[category] += 1

        return name



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

from typing import IO, List, Optional
from urllib import parse

from comet_llm.types import JSONEncodable

from .. import config, constants
from . import comet_api_client, request_exception_wrapper


class ExperimentAPI:
    def __init__(
        self,
        id: str,
        comet_api_client: comet_api_client.CometAPIClient,
        workspace: Optional[str] = None,
        project_name: Optional[str] = None,
    ):
        self._id = id
        self._client = comet_api_client
        self._workspace = workspace
        self._project_name = project_name
        if project_name is None or workspace is None:
            self._project_url = None
        else:
            self._project_url = self._build_comet_url()

    @classmethod
    @request_exception_wrapper.wrap(check_on_prem=True)
    def create_new(  # type: ignore
        cls,
        api_key: str,
        workspace: Optional[str] = None,
        project_name: Optional[str] = None,
    ):
        client = comet_api_client.get(api_key)
        response = client.create_experiment("LLM", workspace, project_name)

        experiment_api = cls(
            id=response[constants.EXPERIMENT_KEY_RESPONSE_KEY],
            comet_api_client=client,
            workspace=response[constants.WORKSPACE_RESPONSE_KEY],
            project_name=response[constants.PROJECT_NAME_RESPONSE_KEY],
        )

        return experiment_api

    @classmethod
    def from_existing_id(  # type: ignore
        cls, id: str, api_key: str, load_metadata: bool = True
    ):
        client = comet_api_client.get(api_key)
        experiment_api = cls(id=id, comet_api_client=client)

        if load_metadata:
            experiment_api.load_metadata()

        return experiment_api

    @property
    def project_url(self) -> Optional[str]:
        return self._project_url

    @property
    def id(self) -> str:
        return self._id

    @property
    def workspace(self) -> Optional[str]:
        return self._workspace

    @property
    def project_name(self) -> Optional[str]:
        return self._project_name

    def load_metadata(self) -> None:
        metadata = self._client.get_experiment_metadata(self._id)
        self._workspace = metadata[constants.WORKSPACE_RESPONSE_KEY]
        self._project_name = metadata[constants.PROJECT_NAME_RESPONSE_KEY]
        self._project_url = self._build_comet_url()

    def _build_comet_url(self) -> str:
        parsed_url = parse.urlparse(config.comet_url())
        parsed_comet_url = f"{parsed_url.scheme}://{parsed_url.netloc}"
        return f"{parsed_comet_url}/{self._workspace}/{self._project_name}"

    @request_exception_wrapper.wrap(check_on_prem=True)
    def log_asset_with_io(self, name: str, file: IO, asset_type: str) -> None:
        self._client.log_experiment_asset_with_io(
            self._id, name=name, file=file, asset_type=asset_type
        )

    @request_exception_wrapper.wrap()
    def log_parameter(self, name: str, value: JSONEncodable) -> None:
        self._client.log_experiment_parameter(self._id, name=name, value=value)

    @request_exception_wrapper.wrap()
    def log_metric(self, name: str, value: JSONEncodable) -> None:
        self._client.log_experiment_metric(self._id, name=name, value=value)

    @request_exception_wrapper.wrap()
    def log_tags(self, tags: List[str]) -> None:
        self._client.log_experiment_tags(self._id, tags=tags)

    @request_exception_wrapper.wrap()
    def log_other(self, name: str, value: JSONEncodable) -> None:
        self._client.log_experiment_other(self._id, name=name, value=value)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

FAILED_TO_SEND_DATA_TO_SERVER = "Failed to send data to server"

UNABLE_TO_LOG_TO_NON_LLM_PROJECT = "Failed to send prompt to the specified project as it is not an LLM project, please specify a different project name."

API_KEY_NOT_FOUND_MESSAGE = """
    CometLLM requires an API key. Please provide it as the
    api_key argument to %s or as an environment
    variable named COMET_API_KEY
    """

NON_ALLOWED_SCORE = "Score can only be 0 or 1 when calling 'log_user_feedback'"

METADATA_KEY_COLLISION_DURING_DEEPMERGE = (
    "Chain or prompt metadata value for the sub-key '%s' was overwritten from '%s' to '%s' during the deep merge",
)

INVALID_TIMESTAMP = "Invalid timestamp: %s. Timestamp must be in seconds if specified."

GLOBAL_CHAIN_NOT_INITIALIZED = "Global chain is not initialized for this thread. Initialize it with `comet_llm.start_chain(...)` if you wish to use %s"



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import inspect
import threading
from typing import TYPE_CHECKING, Dict, Optional

from .. import app, config, exceptions
from . import thread_context_registry

if TYPE_CHECKING:  # pragma: no cover
    from . import chain


class State:
    def __init__(self) -> None:
        self._id: int = 0
        self._thread_context_registry = thread_context_registry.ThreadContextRegistry()
        self._lock = threading.Lock()

    def chain_exists(self) -> bool:
        return self._thread_context_registry.get("global-chain") is not None

    @property
    def chain(self) -> "chain.Chain":
        result: "chain.Chain" = self._thread_context_registry.get("global-chain")
        return result

    @chain.setter
    def chain(self, value: "chain.Chain") -> None:
        self._thread_context_registry.add("global-chain", value)

    def new_id(self) -> int:
        with self._lock:
            self._id += 1
            return self._id


_APP_STATE = State()


def global_chain_exists() -> bool:
    return _APP_STATE.chain_exists()


def get_global_chain() -> "chain.Chain":
    return _APP_STATE.chain


def set_global_chain(new_chain: "chain.Chain") -> None:
    _APP_STATE.chain = new_chain


def get_new_id() -> int:
    return _APP_STATE.new_id()



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import comet_llm.config

from ..import_hooks import finder, registry
from .openai import patcher as openai_patcher


def patch() -> None:
    if comet_llm.config.autologging_enabled() and not comet_llm.config.comet_disabled():
        registry_ = registry.Registry()

        openai_patcher.patch(registry_)

        module_finder = finder.CometFinder(registry_)
        module_finder.hook_into_import_system()



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import collections
import logging
import threading
from typing import DefaultDict

LOGGER = logging.getLogger(__name__)


class Summary:
    def __init__(self) -> None:
        self._logs_registry: DefaultDict[str, int] = collections.defaultdict(lambda: 0)
        self._failed = 0
        self._lock = threading.Lock()

    def add_log(self, project_url: str, name: str) -> None:
        with self._lock:
            if len(self._logs_registry) == 0:
                LOGGER.info("%s logged to %s", name.capitalize(), project_url)

            self._logs_registry[project_url] += 1

    def increment_failed(self) -> None:
        with self._lock:
            self._failed += 1

    def print(self) -> None:
        for project, logs_amount in self._logs_registry.items():
            LOGGER.info("%d prompts and chains logged to %s", logs_amount, project)

        if self._failed > 0:
            LOGGER.info(
                "%d prompts and chains were not logged because of errors", self._failed
            )



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

from typing import Any, Dict, Optional

import flatten_dict


def chain_metadata_to_flat_parameters(
    metadata: Optional[Dict[str, Any]]
) -> Dict[str, Any]:
    metadata_parameters = (
        flatten_dict.flatten(metadata, reducer="dot") if metadata is not None else {}
    )

    result = {
        key: value for key, value in metadata_parameters.items() if value is not None
    }

    return result



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************


import logging
from types import ModuleType
from typing import Dict, Optional, Tuple


def _muted_import_comet_ml() -> Tuple[ModuleType, ModuleType]:
    try:
        logging.disable(logging.CRITICAL)
        import comet_ml
        import comet_ml.config

        return comet_ml, comet_ml.config  # type: ignore
    finally:
        pass
        logging.disable(0)


comet_ml, comet_ml_config = _muted_import_comet_ml()


def _extend_comet_ml_config() -> None:
    CONFIG_MAP_EXTENSION = {
        "comet.disable": {"type": int, "default": 0},
        "comet.logging.console": {"type": str, "default": "INFO"},
        "comet.raise_exceptions_on_error": {"type": int, "default": 0},
        "comet.internal.check_tls_certificate": {"type": bool, "default": True},
    }

    comet_ml_config.CONFIG_MAP.update(CONFIG_MAP_EXTENSION)


def workspace() -> Optional[str]:
    return _COMET_ML_CONFIG["comet.workspace"]  # type: ignore


def project_name() -> Optional[str]:
    return _COMET_ML_CONFIG["comet.project_name"]  # type: ignore


def comet_url() -> str:
    return comet_ml.get_backend_address(_COMET_ML_CONFIG)  # type: ignore


def api_key() -> Optional[str]:
    api_key = comet_ml.get_api_key(None, _COMET_ML_CONFIG)
    return api_key  # type: ignore


def logging_level() -> str:
    return _COMET_ML_CONFIG["comet.logging.console"]  # type: ignore


def is_ready() -> bool:
    """
    True if comet API key is set.
    """
    return api_key() is not None


def comet_disabled() -> bool:
    return bool(_COMET_ML_CONFIG["comet.disable"])


def raising_enabled() -> bool:
    return bool(_COMET_ML_CONFIG["comet.raise_exceptions_on_error"])


def logging_available() -> bool:
    if api_key() is None:
        return False

    return True


def autologging_enabled() -> bool:
    return not comet_ml.get_config("comet.disable_auto_logging")  # type: ignore


def tls_verification_enabled() -> bool:
    return _COMET_ML_CONFIG["comet.internal.check_tls_certificate"]  # type: ignore


def init(
    api_key: Optional[str] = None,
    workspace: Optional[str] = None,
    project: Optional[str] = None,
) -> None:
    """
    An easy, safe, interactive way to set and save your settings.

    Will ask for your api_key if not already set. Your
    api_key will not be shown.

    Will save the config to .comet.config file.
    Default location is "~/" (home) or COMET_CONFIG, if set.

    Args:
        api_key: str (optional) comet API key.
        workspace: str (optional) comet workspace to use for logging.
        project: str (optional) project name to create in comet workspace.

    Valid settings include:
    """

    kwargs: Dict[str, Optional[str]] = {
        "api_key": api_key,
        "workspace": workspace,
        "project_name": project,
    }

    kwargs = {key: value for key, value in kwargs.items() if value is not None}

    comet_ml.init(**kwargs)

    global _COMET_ML_CONFIG
    # Recreate the Config object to re-read the config files
    _COMET_ML_CONFIG = comet_ml.get_config()


_extend_comet_ml_config()

_COMET_ML_CONFIG = comet_ml.get_config()



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import functools
import logging
import sys
from typing import Any, Callable

from . import config

_LOG_ONCE_CACHE = set()


def setup() -> None:
    root = logging.getLogger("comet_llm")
    root.setLevel(config.logging_level())

    console_handler = logging.StreamHandler(sys.stdout)
    root.addHandler(console_handler)


def log_once_at_level(  # type: ignore
    logger: logging.Logger, logging_level: int, message: str, *args, **kwargs
) -> None:
    """
    Log the given message once at the given level then at the DEBUG
    level on further calls.

    This is a global log-once-per-session, as opposed to the
    log-once-per-experiment.
    """
    global _LOG_ONCE_CACHE

    if message not in _LOG_ONCE_CACHE:
        _LOG_ONCE_CACHE.add(message)
        logger.log(logging_level, message, *args, **kwargs)
    else:
        logger.debug(message, *args, **kwargs)


def log_message_on_error(  # type: ignore
    logger: logging.Logger,
    logging_level: int,
    message: str,
    *log_args,
    log_once=False,
    **log_kwargs
) -> Callable:
    def decorator(function: Callable) -> Callable:
        @functools.wraps(function)
        def wrapper(*args, **kwargs) -> Any:  # type: ignore
            try:
                return function(*args, **kwargs)
            except Exception:
                if log_once:
                    log_once_at_level(
                        logger, logging_level, message, *log_args, **log_kwargs
                    )
                else:
                    logger.log(logging_level, message, *log_args, **log_kwargs)

                raise

        return wrapper

    return decorator



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

from typing import Any, Dict, Optional

from ..types import JSONEncodable


def call_data_to_dict(
    prompt: JSONEncodable,
    outputs: JSONEncodable,
    metadata: Optional[Dict[str, Any]] = None,
    prompt_template: Optional[JSONEncodable] = None,
    prompt_template_variables: Optional[JSONEncodable] = None,
    start_timestamp: Optional[float] = None,
    end_timestamp: Optional[float] = None,
    duration: Optional[float] = None,
) -> Dict[str, Any]:
    return {
        "id": 1,
        "category": "llm-call",
        "name": "llm-call-1",
        "inputs": {
            "final_prompt": prompt,
            "prompt_template": prompt_template,
            "prompt_template_variables": prompt_template_variables,
        },
        "outputs": {"output": outputs},
        "duration": duration,
        "start_timestamp": start_timestamp,
        "end_timestamp": end_timestamp,
        "parent_ids": [],
        "metadata": metadata,
    }



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import dataclasses
from typing import List

from .types import AfterCallback, AfterExceptionCallback, BeforeCallback


@dataclasses.dataclass
class CallableExtenders:
    before: List[BeforeCallback] = dataclasses.field(default_factory=list)
    after: List[AfterCallback] = dataclasses.field(default_factory=list)
    after_exception: List[AfterExceptionCallback] = dataclasses.field(
        default_factory=list
    )



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import logging
from typing import TYPE_CHECKING, Dict, List, Optional

from comet_llm import logging as comet_logging

from .. import config, datetimes, exceptions, logging_messages
from ..types import JSONEncodable
from . import deepmerge, state

if TYPE_CHECKING:
    from . import chain

LOGGER = logging.getLogger(__name__)


class Span:
    """
    A single unit of Chain that has its own
    context. Spans can be nested, in that case inner one
    exist in the context of the outer(parent) one.
    Outer Span is considered to be a parent for an inner one.
    """

    def __init__(
        self,
        inputs: JSONEncodable,
        category: str,
        name: Optional[str] = None,
        metadata: Optional[Dict[str, JSONEncodable]] = None,
    ):
        """
        Args:
            inputs: JSONEncodable (required), span inputs.
            category: str (required), span category.
            name: str (optional), span name. If not set will be
                generated automatically.
            metadata: Dict[str, JSONEncodable] (optional), span metadata.
        """
        self._inputs = inputs
        self._category = category
        self._metadata = metadata if metadata is not None else {}
        self._outputs: Optional[Dict[str, JSONEncodable]] = None
        self._context: Optional[List[int]] = None
        self._chain: Optional["chain.Chain"] = None

        self._id = state.get_new_id()
        self._name = name if name is not None else "unnamed"
        self._timer = datetimes.Timer()

    def _connect_to_chain(self, chain: "chain.Chain") -> None:
        chain.track_node(self)
        self._context = chain.context.current()
        self._name = (
            self._name
            if self._name != "unnamed"
            else chain.generate_node_name(self._category)
        )
        self._chain = chain

    @property
    def id(self) -> int:  # pragma: no cover
        return self._id

    @property
    def name(self) -> str:  # pragma: no cover
        return self._name

    def __enter__(self) -> "Span":
        chain = state.get_global_chain()

        if chain is None:
            chain_not_initialized_exception = exceptions.CometLLMException(
                logging_messages.GLOBAL_CHAIN_NOT_INITIALIZED % "`Span`"
            )
            if config.raising_enabled():
                raise chain_not_initialized_exception

            comet_logging.log_once_at_level(
                LOGGER, logging.ERROR, str(chain_not_initialized_exception)
            )

            return self

        self.__api__start__(chain)
        return self

    def __api__start__(self, chain: "chain.Chain") -> None:
        self._connect_to_chain(chain)

        self._timer.start()
        self._chain.context.add(self.id)  # type: ignore

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        self.__api__end__()

    def __api__end__(self) -> None:
        if self._chain is not None:
            self._timer.stop()
            self._chain.context.pop()

    def set_outputs(
        self,
        outputs: Dict[str, JSONEncodable],
        metadata: Optional[Dict[str, JSONEncodable]] = None,
    ) -> None:
        """
        Sets outputs to span object.
        Args:
            outputs: Dict[str, JSONEncodable] (required), outputs
            metadata: Dict[str, JSONEncodable] (optional), span metadata.
                If metadata dictionary was passed to __init__ method,
                it will be updated.
        """
        self._outputs = outputs
        if metadata is not None:
            self._metadata = deepmerge.deepmerge(self._metadata, metadata)

    def as_dict(self) -> Dict[str, JSONEncodable]:
        inputs = self._inputs
        outputs = self._outputs

        inputs = inputs if isinstance(inputs, dict) else {"input": inputs}
        outputs = outputs if isinstance(outputs, dict) else {"output": outputs}

        return {
            "id": self._id,
            "category": self._category,
            "name": self._name,
            "inputs": inputs,
            "outputs": outputs,
            "duration": self._timer.duration,
            "start_timestamp": self._timer.start_timestamp,
            "end_timestamp": self._timer.end_timestamp,
            "parent_ids": self._context,
            "metadata": self._metadata,
        }



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

# type: ignore

from typing import Any


def _dummy_callable(*args, **kwargs):
    pass


class DummyClass:
    def __init__(self, *args, **kwargs) -> None:
        pass

    def __setattr__(self, __name: str, __value: Any) -> None:
        pass

    def __getattribute__(self, __name: str) -> None:
        return _dummy_callable

    def __enter__(self) -> "DummyClass":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        pass



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import logging
from typing import Optional

from . import config, exceptions, experiment_info, logging_messages
from .experiment_api import ExperimentAPI

LOGGER = logging.getLogger(__name__)


@exceptions.filter(allow_raising=config.raising_enabled())
def log_user_feedback(id: str, score: float, api_key: Optional[str] = None) -> None:
    """
    Logs user feedback for the provided Prompt or Chain ID. This will
    overwrite any previously set value.

    Args:
        id: str (required) the ID of the Prompt or Chain.
        score: float (required) the feedback score, can be either 0, 0.0, 1, or 1.0.
        api_key: str (optional) comet API key.
    """
    ALLOWED_SCORES = [0.0, 1.0]

    if score not in ALLOWED_SCORES:
        LOGGER.error(logging_messages.NON_ALLOWED_SCORE)
        return

    info = experiment_info.get(
        api_key,
        api_key_not_found_message=logging_messages.API_KEY_NOT_FOUND_MESSAGE
        % "log_user_feedback",
    )
    experiment_api = ExperimentAPI.from_existing_id(
        id=id,
        api_key=info.api_key,  # type: ignore
        load_metadata=False,
    )

    experiment_api.log_metric("user_feedback", score)


def flush() -> None:
    """Flush all data to Comet platform."""
    pass



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import dataclasses
from typing import Optional

from . import config, exceptions

DEFAULT_PROJECT_NAME = "llm-general"


@dataclasses.dataclass
class ExperimentInfo:
    api_key: Optional[str]
    workspace: Optional[str]
    project_name: Optional[str]


def get(
    api_key: Optional[str] = None,
    workspace: Optional[str] = None,
    project_name: Optional[str] = None,
    api_key_not_found_message: Optional[str] = None,
) -> ExperimentInfo:
    api_key = api_key if api_key else config.api_key()

    if api_key is None and api_key_not_found_message is not None:
        raise exceptions.CometLLMException(
            api_key_not_found_message, log_message_once=True
        )

    workspace = workspace if workspace else config.workspace()
    project_name = project_name if project_name else config.project_name()

    project_name = project_name if project_name else DEFAULT_PROJECT_NAME

    return ExperimentInfo(api_key, workspace, project_name)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

import threading
from typing import Any, Dict


class ThreadContextRegistry:
    def __init__(self) -> None:
        self._registry: Dict[str, Any] = {}
        self._lock = threading.Lock()

    def get(self, key: str) -> Any:
        thread_wise_key = _thread_wise_key(key)
        with self._lock:
            if thread_wise_key not in self._registry:
                return None

            return self._registry[thread_wise_key]

    def add(self, key: str, value: Any) -> None:
        thread_wise_key = _thread_wise_key(key)
        with self._lock:
            self._registry[thread_wise_key] = value


def _thread_wise_key(key: str) -> str:
    return f"{key}-{threading.get_ident()}"



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

from typing import List


class Context:
    def __init__(self) -> None:
        self._stack: List[int] = []

    def add(self, span_id: int) -> None:
        self._stack.append(span_id)

    def pop(self) -> None:
        if len(self._stack) > 0:
            self._stack.pop()

    def current(self) -> List[int]:
        return list(self._stack)



# -*- coding: utf-8 -*-
# *******************************************************
#   ____                     _               _
#  / ___|___  _ __ ___   ___| |_   _ __ ___ | |
# | |   / _ \| '_ ` _ \ / _ \ __| | '_ ` _ \| |
# | |__| (_) | | | | | |  __/ |_ _| | | | | | |
#  \____\___/|_| |_| |_|\___|\__(_)_| |_| |_|_|
#
#  Sign up for free at https://www.comet.com
#  Copyright (C) 2015-2023 Comet ML INC
#  This source code is licensed under the MIT license found in the
#  LICENSE file in the root directory of this package.
# *******************************************************

from typing import List, Optional

import comet_ml

from .. import experiment_info, logging_messages, query_dsl
from . import llm_trace_api

# TODO: make the decision about dependencies from comet-ml. Address testing.


class API:
    def __init__(self, api_key: Optional[str] = None) -> None:
        experiment_info_ = experiment_info.get(
            api_key,
            api_key_not_found_message=logging_messages.API_KEY_NOT_FOUND_MESSAGE
            % "API",
        )
        self._api = comet_ml.API(api_key=experiment_info_.api_key, cache=False)

    def get_llm_trace_by_key(self, trace_key: str) -> llm_trace_api.LLMTraceAPI:
        """
        Get an API Trace object by key.

        Args:
            trace_key: str, key of the prompt or chain

        Returns: An LLMTraceAPI object that can be used to get or update trace data
        """
        matching_trace = self._api.get_experiment_by_key(trace_key)

        if matching_trace is None:
            raise ValueError(
                f"Failed to find any matching traces with the key {trace_key}"
            )

        return llm_trace_api.LLMTraceAPI.__api__from_api_experiment__(matching_trace)

    def get_llm_trace_by_name(
        self, workspace: str, project_name: str, trace_name: str
    ) -> llm_trace_api.LLMTraceAPI:
        """
        Get an API Trace object by name.

        Args:
            workspace: str, name of the workspace
            project_name: str, name of the project
            trace_name: str, name of the prompt or chain

        Returns: An LLMTraceAPI object that can be used to get or update trace data
        """
        matching_trace = self._api.query(
            workspace, project_name, query_dsl.Other("Name") == trace_name
        )

        if len(matching_trace) == 0:
            raise ValueError(
                f"Failed to find any matching traces with the name {trace_name} in the project {project_name}"
            )
        elif len(matching_trace) > 1:
            raise ValueError(
                f"Found multiple traces with the name {trace_name} in the project {project_name}"
            )

        return llm_trace_api.LLMTraceAPI.__api__from_api_experiment__(matching_trace[0])

    def query(
        self, workspace: str, project_name: str, query: str
    ) -> List[llm_trace_api.LLMTraceAPI]:
        """
        Fetch LLM Trace based on a query. Currently it is only possible to use
        trace metadata or details fields to filter the traces.

        Args:
            workspace: str, name of the workspace
            project_name: str, name of the project
            query: str, name of the prompt or chain

        Returns: A list of LLMTrace objects

        Notes:
        The `query` object takes the form of (QUERY_VARIABLE OPERATOR VALUE) with:

        * QUERY_VARIABLE is either TraceMetadata, Duration, Timestamp.
        * OPERATOR is any standard mathematical operators `<=`, `>=`, `!=`, `<`, `>`.

        It is also possible to add multiple query conditions using `&`.

        If you are querying nested parameters, you should flatted the parameter name using the
        `.` operator.

        To query the duration, you can use Duration().

        Example:
        ```python
        # Find all traces where the metadata field `token` is greater than 50
        api.query("workspace", "project", TraceMetadata("token") > 50)

        # Find all traces where the duration field is between 1 second and 2 seconds
        api.query("workspace", "project", (Duration() > 1) & (Duration() <= 2))

        # Find all traces based on the timestamp
        api.query("workspace", "project", Timestamp() > datetime(2023, 9, 10))

        ```
        """
        matching_api_objects = self._api.query(workspace, project_name, query)

        return [
            llm_trace_api.LLMTraceAPI.__api__from_api_experiment__(api_object)
            for api_object in matching_api_objects
        ]
